from .author_card import AuthorCard as AuthorCard
from .blog_header import BlogHeader as BlogHeader
from .blog_major import BlogMajor as BlogMajor
from .blog_minor import BlogMinor as BlogMinor
from .blog_section import BlogSection as BlogSection
from .bullet_point import BulletPoint as BulletPoint
from .card_section import CardSection as CardSection
from .custom_switcher_bar import CustomSwitcherBar as CustomSwitcherBar
from .fake_link import FakeLink as FakeLink
from .faq import Faq as Faq
from .faq_section import FaqSection as FaqSection
from .footer import Footer as Footer
from .get_started import GetStarted as GetStarted
from .goody_column import GoodyColumn as GoodyColumn
from .hero_card import HeroCard as HeroCard
from .hero_section import HeroSection as HeroSection
from .image_placeholder import ImagePlaceholder as ImagePlaceholder
from .major_column import MajorColumn as MajorColumn
from .navbar import Navbar as Navbar
from .outlined_colored_button import (
    OutlinedColoredButton as OutlinedColoredButton,
)
from .outlined_neutral_button import (
    OutlinedNeutralButton as OutlinedNeutralButton,
)
from .pricing_cards import PricingCards as PricingCards
from .pricing_section import PricingSection as PricingSection
from .root_component import RootComponent as RootComponent
from .service_power import ServicePower as ServicePower
from .service_speed import ServiceSpeed as ServiceSpeed
from .subscribe_input_button import SubscribeInputButton as SubscribeInputButton
from .testimonial import Testimonial as Testimonial
from .testimonials import Testimonials as Testimonials
